from os import environ

from ..concurrency import WorkerPool
from ..lock import softlock_add, softlock_list, softlock_remove
from ..utils.cmdline import get_target_nodes
from ..utils.table import ROW_SEPARATOR, render_table
from ..utils.text import (
    bold,
    error_summary,
    format_timestamp,
    green,
    mark_for_translation as _,
    randstr,
    red,
    yellow,
)
from ..utils.ui import io, page_lines


def remove_dummy_nodes(targets):
    _targets = []
    for node in targets:
        if node.dummy:
            io.stdout(_("{x} {node}  is a dummy node").format(node=bold(node.name), x=yellow("»")))
        else:
            _targets.append(node)
    return _targets


def remove_lock_if_present(node, lock_id):
    for lock in softlock_list(node):
        if lock['id'] == lock_id:
            softlock_remove(node, lock_id)
            return True
    return False


def bw_lock_add(repo, args):
    errors = []
    target_nodes = get_target_nodes(repo, args['targets'])
    target_nodes = remove_dummy_nodes(target_nodes)
    pending_nodes = target_nodes[:]
    max_node_name_length = max([len(node.name) for node in target_nodes])
    lock_id = randstr(length=4).upper()
    io.progress_set_total(len(pending_nodes))

    def tasks_available():
        return bool(pending_nodes)

    def next_task():
        node = pending_nodes.pop()
        return {
            'target': softlock_add,
            'task_id': node.name,
            'args': (node, lock_id),
            'kwargs': {
                'comment': args['comment'],
                'expiry': args['expiry'],
                'item_selectors': args['items'],
            },
        }

    def handle_result(task_id, return_value, duration):
        io.progress_advance()
        io.stdout(_("{x} {node}  locked with ID {id} (expires in {exp})").format(
            x=green("✓"),
            node=bold(task_id.ljust(max_node_name_length)),
            id=return_value,
            exp=args['expiry'],
        ))

    def handle_exception(task_id, exception, traceback):
        msg = "{}: {}".format(task_id, exception)
        io.stderr(traceback)
        io.stderr(repr(exception))
        io.stderr(msg)
        errors.append(msg)

    worker_pool = WorkerPool(
        tasks_available,
        next_task,
        handle_exception=handle_exception,
        handle_result=handle_result,
        pool_id="lock",
        workers=args['node_workers'],
    )
    worker_pool.run()

    error_summary(errors)


def bw_lock_remove(repo, args):
    errors = []
    target_nodes = get_target_nodes(repo, args['targets'])
    target_nodes = remove_dummy_nodes(target_nodes)
    pending_nodes = target_nodes[:]
    max_node_name_length = max([len(node.name) for node in target_nodes])
    io.progress_set_total(len(pending_nodes))

    def tasks_available():
        return bool(pending_nodes)

    def next_task():
        node = pending_nodes.pop()
        return {
            'target': remove_lock_if_present,
            'task_id': node.name,
            'args': (node, args['lock_id'].upper()),
        }

    def handle_result(task_id, return_value, duration):
        io.progress_advance()
        if return_value is True:
            io.stdout(_("{x} {node}  lock {id} removed").format(
                x=green("✓"),
                node=bold(task_id.ljust(max_node_name_length)),
                id=args['lock_id'].upper(),
            ))
        else:
            io.stderr(_(
                "{x} {node}  has no lock with ID {id}"
            ).format(
                x=red("!"),
                node=bold(task_id.ljust(max_node_name_length)),
                id=args['lock_id'].upper(),
            ))

    def handle_exception(task_id, exception, traceback):
        msg = "{}: {}".format(task_id, exception)
        io.stderr(traceback)
        io.stderr(repr(exception))
        io.stderr(msg)
        errors.append(msg)

    worker_pool = WorkerPool(
        tasks_available,
        next_task,
        handle_exception=handle_exception,
        handle_result=handle_result,
        pool_id="lock_remove",
        workers=args['node_workers'],
    )
    worker_pool.run()

    error_summary(errors)


def bw_lock_show(repo, args):
    errors = []
    target_nodes = get_target_nodes(repo, args['targets'])
    target_nodes = remove_dummy_nodes(target_nodes)
    pending_nodes = target_nodes[:]
    locks_on_node = {}

    def tasks_available():
        return bool(pending_nodes)

    def next_task():
        node = pending_nodes.pop()
        return {
            'target': softlock_list,
            'task_id': node.name,
            'args': (node,),
        }

    def handle_result(task_id, return_value, duration):
        locks_on_node[task_id] = return_value
        repo.hooks.lock_show(repo, repo.get_node(task_id), return_value)

    def handle_exception(task_id, exception, traceback):
        msg = "{}: {}".format(task_id, exception)
        io.stderr(traceback)
        io.stderr(repr(exception))
        io.stderr(msg)
        errors.append(msg)

    worker_pool = WorkerPool(
        tasks_available,
        next_task,
        handle_exception=handle_exception,
        handle_result=handle_result,
        pool_id="lock_show",
        workers=args['node_workers'],
    )
    worker_pool.run()

    if errors:
        error_summary(errors)
        return

    rows = [[
        bold(_("node")),
        bold(_("ID")),
        bold(_("created")),
        bold(_("expires")),
        bold(_("user")),
        bold(_("items")),
        bold(_("comment")),
    ], ROW_SEPARATOR]

    for node_name, locks in sorted(locks_on_node.items()):
        if locks:
            first_lock = True
            for lock in locks:
                lock['formatted_date'] = format_timestamp(lock['date'])
                lock['formatted_expiry'] = format_timestamp(lock['expiry'])
                first_item = True
                for item in lock['items']:
                    rows.append([
                        node_name if first_item and first_lock else "",
                        lock['id'] if first_item else "",
                        lock['formatted_date'] if first_item else "",
                        lock['formatted_expiry'] if first_item else "",
                        lock['user'] if first_item else "",
                        item,
                        lock['comment'] if first_item else "",
                    ])
                    # always repeat for grep style
                    first_item = environ.get("BW_TABLE_STYLE") == 'grep'
                # always repeat for grep style
                first_lock = environ.get("BW_TABLE_STYLE") == 'grep'
        else:
            rows.append([
                node_name,
                _("(none)"),
                "",
                "",
                "",
                "",
                "",
            ])
        rows.append(ROW_SEPARATOR)

    page_lines(render_table(
        rows[:-1],  # remove trailing ROW_SEPARATOR
        alignments={1: 'center'},
    ))



import platform
from subprocess import Popen, PIPE

from ..bundle import FILENAME_BUNDLE
from ..secrets import FILENAME_SECRETS


HOST_OS = {
    "Darwin": 'macos',
    "Linux": 'linux',
}


def host_os():
    return HOST_OS[platform.system()]


def make_repo(tmpdir, bundles=None, groups=None, nodes=None):
    bundles = {} if bundles is None else bundles
    groups = {} if groups is None else groups
    nodes = {} if nodes is None else nodes

    bundles_dir = tmpdir.mkdir("bundles")
    for bundle, items in bundles.items():
        bundle_dir = bundles_dir.mkdir(bundle)
        bundle_dir.mkdir("files")
        bundlepy = bundle_dir.join(FILENAME_BUNDLE)
        bundle_content = "# -*- coding: utf-8 -*-\n"
        for itemtype, itemconfig in items.items():
            bundle_content += "{} = {}\n".format(itemtype, repr(itemconfig))
        bundlepy.write(bundle_content)

    tmpdir.mkdir("data")
    tmpdir.mkdir("hooks")
    tmpdir.mkdir("libs")

    groupspy = tmpdir.join("groups.py")
    groupspy.write("groups = {}\n".format(repr(groups)))

    nodespy = tmpdir.join("nodes.py")
    nodespy.write("nodes = {}\n".format(repr(nodes)))

    secrets = tmpdir.join(FILENAME_SECRETS)
    secrets.write("[generate]\nkey = {}\n\n[encrypt]\nkey = {}\n".format(
        "Fl53iG1czBcaAPOKhSiJE7RjFU9nIAGkiKDy0k_LoTc=",
        "DbYiUu5VMfrdeSiKYiAH4rDOAUISipvLSBJI-T0SpeY=",
    ))


def run(command, path=None):
    process = Popen(command, cwd=path, shell=True, stderr=PIPE, stdout=PIPE)
    stdout, stderr = process.communicate()
    print(stdout.decode('utf-8'))
    print(stderr.decode('utf-8'))
    return (stdout, stderr, process.returncode)



from shlex import quote

from bundlewrap.exceptions import BundleError
from bundlewrap.items import Item
from bundlewrap.utils.text import mark_for_translation as _


def svc_start(node, svcname):
    return node.run("initctl start --no-wait -- {}".format(quote(svcname)), may_fail=True)


def svc_running(node, svcname):
    result = node.run("initctl status -- {}".format(quote(svcname)), may_fail=True)
    if result.return_code != 0:
        return False
    return " start/" in result.stdout_text


def svc_stop(node, svcname):
    return node.run("initctl stop --no-wait -- {}".format(quote(svcname)), may_fail=True)


class SvcUpstart(Item):
    """
    A service managed by Upstart.
    """
    BUNDLE_ATTRIBUTE_NAME = "svc_upstart"
    ITEM_ATTRIBUTES = {
        'running': True,
    }
    ITEM_TYPE_NAME = "svc_upstart"

    def __repr__(self):
        return "<SvcUpstart name:{} running:{}>".format(
            self.name,
            self.attributes['running'],
        )

    def fix(self, status):
        if self.attributes['running'] is False:
            svc_stop(self.node, self.name)
        else:
            svc_start(self.node, self.name)

    def get_canned_actions(self):
        return {
            'reload': {
                'command': "reload {}".format(self.name),
                'needs': [self.id],
            },
            'restart': {
                'command': "restart {}".format(self.name),
                'needs': [self.id],
            },
            'stopstart': {
                'command': "stop {0} && start {0}".format(self.name),
                'needs': [self.id],
            },
        }

    def sdict(self):
        return {'running': svc_running(self.node, self.name)}

    @classmethod
    def validate_attributes(cls, bundle, item_id, attributes):
        if not isinstance(attributes.get('running', True), bool):
            raise BundleError(_(
                "expected boolean for 'running' on {item} in bundle '{bundle}'"
            ).format(
                bundle=bundle.name,
                item=item_id,
            ))



from sys import exit

from ..exceptions import NoSuchGroup, NoSuchItem, NoSuchNode
from . import names
from .text import mark_for_translation as _, red
from .ui import io, QUIT_EVENT


def count_items(nodes):
    count = 0
    for node in nodes:
        if QUIT_EVENT.is_set():
            return 0
        count += len(node.items)
    return count


def get_group(repo, group_name):
    try:
        return repo.get_group(group_name)
    except NoSuchGroup:
        io.stderr(_("{x} No such group: {group}").format(
            group=group_name,
            x=red("!!!"),
        ))
        exit(1)


def get_item(node, item_id):
    try:
        return node.get_item(item_id)
    except NoSuchItem:
        io.stderr(_("{x} No such item on node '{node}': {item}").format(
            item=item_id,
            node=node.name,
            x=red("!!!"),
        ))
        exit(1)


def get_node(repo, node_name):
    try:
        return repo.get_node(node_name)
    except NoSuchNode:
        io.stderr(_("{x} No such node: {node}").format(
            node=node_name,
            x=red("!!!"),
        ))
        exit(1)


HELP_get_target_nodes = _("""expression to select target nodes:

my_node            # to select a single node
my_group           # all nodes in this group
bundle:my_bundle   # all nodes with this bundle
!bundle:my_bundle  # all nodes without this bundle
!group:my_group    # all nodes not in this group
"lambda:node.metadata_get('foo/magic', 47) < 3"
                   # all nodes whose metadata["foo"]["magic"] is less than three
""")


def get_target_nodes(repo, target_strings):
    targets = set()
    for name in target_strings:
        name = name.strip()
        if name.startswith("bundle:"):
            bundle_name = name.split(":", 1)[1]
            for node in repo.nodes:
                if bundle_name in names(node.bundles):
                    targets.add(node)
        elif name.startswith("!bundle:"):
            bundle_name = name.split(":", 1)[1]
            for node in repo.nodes:
                if bundle_name not in names(node.bundles):
                    targets.add(node)
        elif name.startswith("!group:"):
            group_name = name.split(":", 1)[1]
            for node in repo.nodes:
                if group_name not in names(node.groups):
                    targets.add(node)
        elif name.startswith("lambda:"):
            expression = eval("lambda node: " + name.split(":", 1)[1])
            for node in repo.nodes:
                if expression(node):
                    targets.add(node)
        else:
            try:
                targets.add(repo.get_node(name))
            except NoSuchNode:
                try:
                    group = repo.get_group(name)
                except NoSuchGroup:
                    io.stderr(_("{x} No such node or group: {name}").format(
                        x=red("!!!"),
                        name=name,
                    ))
                    exit(1)
                else:
                    targets.update(group.nodes)
    return targets



from base64 import b64encode
from codecs import getwriter
from contextlib import contextmanager
import hashlib
from inspect import isgenerator
from os import chmod, close, makedirs, remove
from os.path import dirname, exists
from random import shuffle
import stat
from sys import stderr, stdout
from tempfile import mkstemp

from requests import get

from ..exceptions import DontCache, FaultUnavailable


class NO_DEFAULT: pass
MODE644 = stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH
STDERR_WRITER = getwriter('utf-8')(stderr.buffer)
STDOUT_WRITER = getwriter('utf-8')(stdout.buffer)


def cached_property(prop):
    """
    A replacement for the property decorator that will only compute the
    attribute's value on the first call and serve a cached copy from
    then on.
    """
    def cache_wrapper(self):
        if not hasattr(self, "_cache"):
            self._cache = {}
        if prop.__name__ not in self._cache:
            try:
                return_value = prop(self)
                if isgenerator(return_value):
                    return_value = tuple(return_value)
            except DontCache as exc:
                return exc.obj
            else:
                self._cache[prop.__name__] = return_value
        return self._cache[prop.__name__]
    return property(cache_wrapper)


def download(url, path):
    with error_context(url=url, path=path):
        if not exists(dirname(path)):
            makedirs(dirname(path))
        if exists(path):
            chmod(path, MODE644)
        with open(path, 'wb') as f:
            r = get(url, stream=True)
            r.raise_for_status()
            for block in r.iter_content(1024):
                if not block:
                    break
                else:
                    f.write(block)


class ErrorContext(Exception):
    pass


@contextmanager
def error_context(**kwargs):
    """
    This can be used to provide context for critical exceptions. Since
    we're processing lots of different dicts, a "KeyError: foo" will
    often not be helpful, since it's not clear which dict is missing the
    key.


    >>> with error_context(arbitrary_kwarg="helpful hint"):
    ...     {}["foo"]
    ...
    Traceback (most recent call last):
      [...]
    KeyError: 'foo'

    The above exception was the direct cause of the following exception:

    Traceback (most recent call last):
      [...]
    bundlewrap.utils.ErrorContext: ACTUAL EXCEPTION ABOVE
    {'arbitrary_kwarg': 'helpful hint'}


    Careful though: Only use this in places where you don't expect
    exceptions to occur, since they will indiscriminately be reraised as
    ErrorContext.
    """
    try:
        yield
    except Exception as exc:
        raise ErrorContext("ACTUAL EXCEPTION ABOVE\n" + repr(kwargs)) from exc


class Fault:
    """
    A proxy object for lazy access to things that may not really be
    available at the time of use.

    This let's us gracefully skip items that require information that's
    currently not available.
    """
    def __init__(self, fault_identifier, callback, **kwargs):
        if isinstance(fault_identifier, list):
            self.id_list = fault_identifier
        else:
            self.id_list = [fault_identifier]

        for key, value in sorted(kwargs.items()):
            self.id_list.append(hash(key))
            self.id_list.append(hash(value))

        self._available = None
        self._exc = None
        self._value = None
        self.callback = callback
        self.kwargs = kwargs

    def _resolve(self):
        if self._available is None:
            try:
                self._value = self.callback(**self.kwargs)
                if isinstance(self._value, Fault):
                    self._value = self._value.value
                self._available = True
            except FaultUnavailable as exc:
                self._available = False
                self._exc = exc

    def __add__(self, other):
        if isinstance(other, Fault):
            def callback():
                return self.value + other.value
            return Fault(self.id_list + other.id_list, callback)
        else:
            def callback():
                return self.value + other
            return Fault(self.id_list + ['raw {}'.format(repr(other))], callback)

    def __eq__(self, other):
        if not isinstance(other, Fault):
            return False
        else:
            return self.id_list == other.id_list

    def __hash__(self):
        return hash(tuple(self.id_list))

    def __len__(self):
        return len(self.value)

    def __lt__(self, other):
        return self.value < other.value

    def __str__(self):
        return str(self.value)

    def b64encode(self):
        def callback():
            return b64encode(self.value.encode('UTF-8')).decode('UTF-8')
        return Fault(self.id_list + ['b64encode'], callback)

    def format_into(self, format_string):
        def callback():
            return format_string.format(self.value)
        return Fault(self.id_list + ['format_into ' + format_string], callback)

    @property
    def is_available(self):
        self._resolve()
        return self._available

    @property
    def value(self):
        self._resolve()
        if not self._available:
            raise self._exc
        return self._value


def _make_method_callback(method_name):
    def method(self, *args, **kwargs):
        def callback():
            return getattr(self.value, method_name)(*args, **kwargs)
        return Fault(self.id_list + [method_name], callback)
    return method


for method_name in (
    'format',
    'lower',
    'lstrip',
    'replace',
    'rstrip',
    'strip',
    'upper',
    'zfill',
):
    setattr(Fault, method_name, _make_method_callback(method_name))


def get_file_contents(path):
    with error_context(path=path):
        with open(path, 'rb') as f:
            content = f.read()
    return content


def hash_local_file(path):
    """
    Retuns the sha1 hash of a file on the local machine.
    """
    return sha1(get_file_contents(path))


def names(obj_list):
    """
    Iterator over the name properties of a given list of objects.

    repo.nodes          will give you node objects
    names(repo.nodes)   will give you node names
    """
    for obj in obj_list:
        yield obj.name


def randomize_order(obj):
    if isinstance(obj, dict):
        result = list(obj.items())
    else:
        result = list(obj)
    shuffle(result)
    return result


def sha1(data):
    """
    Returns hex SHA1 hash for input.
    """
    hasher = hashlib.sha1()
    hasher.update(data)
    return hasher.hexdigest()


class SkipList:
    """
    Used to maintain a list of nodes that have already been visited.
    """
    def __init__(self, path):
        self.path = path
        if path and exists(path):
            with open(path) as f:
                self._list_items = set(f.read().strip().split("\n"))
        else:
            self._list_items = set()

    def __contains__(self, item):
        return item in self._list_items

    def add(self, item):
        if self.path:
            self._list_items.add(item)

    def dump(self):
        if self.path:
            with open(self.path, 'w') as f:
                f.write("\n".join(sorted(self._list_items)) + "\n")


@contextmanager
def tempfile():
    handle, path = mkstemp()
    close(handle)
    yield path
    remove(path)



from shlex import quote
from subprocess import CalledProcessError, check_output, STDOUT

from .text import mark_for_translation as _


def get_git_branch():
    try:
        return check_output(
            "git rev-parse --abbrev-ref HEAD",
            shell=True,
            stderr=STDOUT,
        ).strip()
    except CalledProcessError:
        return None


def get_git_clean():
    try:
        return not bool(check_output(
            "git status --porcelain",
            shell=True,
            stderr=STDOUT,
        ).strip())
    except CalledProcessError:
        return None


def get_bzr_rev():
    try:
        return check_output(
            "bzr revno",
            shell=True,
            stderr=STDOUT,
        ).strip()
    except CalledProcessError:
        return None


def get_git_rev():
    try:
        return check_output(
            "git rev-parse HEAD",
            shell=True,
            stderr=STDOUT,
        ).strip()
    except CalledProcessError:
        return None


def get_hg_rev():
    try:
        return check_output(
            "hg --debug id -i",
            shell=True,
            stderr=STDOUT,
        ).strip().rstrip("+")
    except CalledProcessError:
        return None


def get_rev():
    for scm_rev in (get_git_rev, get_hg_rev, get_bzr_rev):
        rev = scm_rev()
        if rev is not None:
            return rev
    return None


def set_git_rev(rev, detach=False):
    if not get_git_clean():
        raise RuntimeError(_("git working dir not clean, won't change rev"))
    if detach:
        command = "git checkout --detach {}".format(quote(rev))
    else:
        command = "git checkout {}".format(quote(rev))
    check_output(
        command,
        shell=True,
        stderr=STDOUT,
    )
